//   Copyright 2007, 2008 Nicolas Dubuit
//
//     This file is part of Immix.
//
//     Immix is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
using namespace std ;

#ifdef Q_OS_WIN32
  #include <exiv2\exif.hpp>
  #include <exiv2\image.hpp>
#else
  #include <exiv2/exif.hpp>
  #include <exiv2/image.hpp>
#endif

#include <QCoreApplication>
#include <QApplication>
#include <QFileInfo>

#include <QPainter>
#include <QFileDialog>
#include <QLabel>
#include <QImage>
#include <QVBoxLayout>
#include <QMouseEvent>
#include <QMatrix>
#include <QProgressDialog>
#include <QMessageBox>

#include "immix.h"
#include "QNImageDialog.h"
#include "initialwindow.h"
#include "align.h"

int ImmixWin::MkrSz = 200 ;
int ImmixWin::FineFactor = 1 ;
int ImmixWin::jpgQuality = 100 ;
float ImmixWin::ZoomFactor = 1. ;

int ImmixDisplay::margin = 7 ;

QString ImmixWin::Suffix ("__immix.jpg");




ImmixWin::ImmixWin (int& err, QWidget* parent)
	: QStackedWidget (parent)
{	
// No marker.
	mkr1 = QPoint (-1,-1) ;
	mkr2 = QPoint (-1,-1) ;

// Create alignment manager.
    alignManager = new AlignmentManager (this) ;

// Create initial window.
    InitialWindow* initWin = new InitialWindow (this, QString(":/immix.svg")) ;
    QWidget* principalWin = new QWidget () ;

// Create principal window. 
	stack = new QStackedWidget () ;
	slider = new QSlider (Qt::Horizontal) ;
	slider->setRange (0, 0) ;
	slider->setTickInterval (1) ;
	slider->setTickPosition (QSlider::TicksBelow) ;
	slider->setTracking (false) ;
    slider->setToolTip (tr("Change the displayed file.")) ;
    slider->setFocus (Qt::OtherFocusReason) ;
    labFile = new QLabel () ;
	btnAlign = new QPushButton (QIcon(":/icons/align.svg"), tr("Align")) ;
	btnAlign->setEnabled (false) ;
    btnAlign->setToolTip (tr("Align all images on the first one.")) ;
	btnAverage = new QPushButton (QIcon(":/icons/merge.svg"),tr("Average")) ;
    btnAverage->setToolTip (tr("Average images.<i>Images should be aligned first.</i>")) ;
	btnRemove = new QPushButton (QIcon(":/icons/list-remove.svg"),tr("Remove")) ;
    btnRemove->setToolTip (tr("Remove current image from file list.")) ;
	btnRevert = new QPushButton (QIcon(":/icons/edit-undo.svg"),tr("Revert")) ;
	btnRevert->setEnabled (false) ;
    btnRevert->setToolTip (tr("Remove aligned images and reload initial files.")) ;
	spnMkr = new QSpinBox () ;
	spnMkr->setRange (50, 800) ;
	spnMkr->setValue (MkrSz) ;
    spnMkr->setSuffix (" px") ;
    spnMkr->setToolTip (tr("Change marker size. <i>The marker size must be larger than the maximum shift between images. </i>")) ;
	spnFineFact = new QSpinBox () ;
	spnFineFact->setRange (1, 5) ;
	spnFineFact->setValue (FineFactor) ;
    spnFineFact->setSuffix("x") ;
    spnFineFact->setToolTip (tr("Alignment (subpixel) precision. <i>A value of 2 indicates that the alignment precision should be 1/2 pixel. Values higher than 3 induce high memory consumption and few improvement.</i>")) ;
	btnQuit = new QPushButton (tr("Quit")) ;

// Window layout.
    QVBoxLayout *layout = new QVBoxLayout (principalWin) ;
    layout->addWidget (new QLabel (tr("Left click : set 1st marker ; right click : set 2nd marker."))) ;
    layout->addWidget(stack);
    QHBoxLayout *sliderLayout = new QHBoxLayout () ;
    sliderLayout->addWidget(slider) ;
    sliderLayout->addWidget(labFile) ;
    layout->addLayout(sliderLayout);
    QHBoxLayout *btnLayout = new QHBoxLayout () ;
    btnLayout->addWidget(btnAlign) ;
    btnLayout->addWidget(btnAverage) ;
    btnLayout->addSpacing (10) ;
    btnLayout->addWidget(btnRemove) ;
    btnLayout->addSpacing (10) ;
    btnLayout->addWidget(btnRevert) ;
    btnLayout->addStretch() ;
    btnLayout->addWidget(new QLabel(tr("Size :"))) ;
    btnLayout->addWidget(spnMkr) ;
    btnLayout->addWidget(new QLabel(tr("FineFactor :"))) ;
    btnLayout->addWidget(spnFineFact) ;
    btnLayout->addStretch() ;
    btnLayout->addWidget(btnQuit) ;
    layout->addLayout(btnLayout);

// This beeing a stacked widget, let's add the main windows to it.
    addWidget (initWin) ;
    addWidget (principalWin) ;
    setCurrentIndex (0) ;

// Connect signals.
    connect (slider, SIGNAL(valueChanged(int)),
             this, SLOT(SwitchTo(int)) ) ;
    connect (btnAlign, SIGNAL(clicked()),
             alignManager, SLOT(startAlign()) ) ;
    connect (btnAverage, SIGNAL(clicked()),
             this, SLOT(StartAverage()) ) ;
    connect (btnRemove, SIGNAL(clicked()),
             this, SLOT(RemoveCurrent()) ) ;
    connect (btnRevert, SIGNAL(clicked()),
             this, SLOT(Revert()) ) ;
    connect (btnQuit, SIGNAL(clicked()),
             QCoreApplication::instance(), SLOT(quit()) ) ;
    connect (spnMkr, SIGNAL(valueChanged(int)),
             this, SLOT (ChangeMarkerSize(int))) ;
    connect (spnFineFact, SIGNAL(valueChanged(int)),
             this, SLOT (ChangeFineFactor(int))) ;

// Try to get the list of files from command line.
    QStringList files = qApp->arguments() ;
    files.removeFirst () ;
    if (files.size() >= 2)
        SetFileList (files) ;

// No problem.
    err = 0;
    show () ;

// Define the size of the principal window.
    setMinimumSize ( 2*int(ImmixWin::ZoomFactor*MkrSz) + 20, int(ImmixWin::ZoomFactor*MkrSz) + 400) ;
}


void ImmixWin::SetFileList (const QStringList& files)
{
// Check file list length.
    if (files.size()<2)
        return ;

// Read file list.
    QString file ;
    foreach (file, files)
        AddFile (file) ;

// Initialize slider.
    stack->setCurrentIndex (0) ;
    displayList[0]->Load() ;
    slider->setValue (0) ;
    activeIndex = 0 ;
    labFile->setText ( QFileInfo(displayList[0]->filePath[0]).fileName() ) ;

// Show principal window.
    setCurrentIndex (1) ;
}

void ImmixWin::AddFile (const QString& file)
{
	QPointer<ImmixDisplay> displPtr = new ImmixDisplay (this, file) ;
	stack->addWidget (displPtr) ;
	displayList.append (displPtr) ;
	slider->setRange (0, displayList.size()-1) ;
    displPtr->setFocusProxy(slider) ;
}

void ImmixWin::ChangeMarkerSize (int newSize)
{
	MkrSz = newSize ;
	update() ;
}

void ImmixWin::ChangeFineFactor (int newFineFactor)
{
    if ( (newFineFactor > 3 ) && (FineFactor <= 3) )
        if ( QMessageBox::No == QMessageBox::warning (this, "immix", tr("Fine factor higher than 2 are usually\nnot useful and may need a themendous amount of memory.\n\nIncrease fine factor anyway?"), QMessageBox::Yes | QMessageBox::No) )
        {
            spnFineFact->setValue (FineFactor) ;
            return ;
        }

    FineFactor = newFineFactor ;
}


void ImmixWin::RemoveCurrent (void)
{
	// This program is useless for only 1 image
	if (displayList.size()<3) return ;

	// The first image is always loaded.
	if ( activeIndex == 0 )
		displayList[1]->Load() ;

	// Choose a new active index.
	int newActiveIndex ;
	if ( activeIndex != displayList.size()-1 )
		newActiveIndex = activeIndex + 1 ;
	else
		newActiveIndex = activeIndex - 1 ;
	stack->setCurrentIndex (newActiveIndex) ;

	// Delete the element.
	displayList[activeIndex]->Unload () ;
	displayList[newActiveIndex]->Load () ;
	delete displayList[activeIndex] ;
	displayList.removeAt (activeIndex) ;
	activeIndex = newActiveIndex ;

	// Shrink the slider.
	slider->setMaximum (displayList.size()-1) ;

    // Give input focus to slider.
    slider->setFocus (Qt::OtherFocusReason) ;

	update() ;
}



void ImmixWin::StartAverage (void) 
{
	// First we prepare the tables. We get the size from the (always-loaded) first image.
	// To limit memory usage, we could use unsigned shorts. This is OK with up to 256 images.
	int width = (displayList[0]->imgPtr)->width() ;
	int height = (displayList[0]->imgPtr)->height() ;
	unsigned long * red = new unsigned long [width * height] ;
	unsigned long * green = new unsigned long [width * height] ;
	unsigned long * blue = new unsigned long [width * height] ;
	if ( ! (red && green && blue) )
	{
		QMessageBox::critical(this, tr("Memory error"), tr("Not enough memory guys!") ) ;
		if (red) delete[] red ;
		if (green) delete[] green ;
		if (blue) delete[] blue ;
		return ;
	}

	// Disable most actions.
	spnMkr->setEnabled (false) ;
	btnAlign->setEnabled (false) ;
	btnAverage->setEnabled (false) ;
	btnRemove->setEnabled (false) ;
	btnRevert->setEnabled (false) ;
	btnQuit->setEnabled (false) ;

	// Declaration of loop indexes.
	int i,x,y ;
	QImage img ;
	QRgb clr ;

	// We initialize the data.
    memset(red, 0, width * height * sizeof(unsigned long));
    memset(green, 0, width * height * sizeof(unsigned long));
    memset(blue, 0, width * height * sizeof(unsigned long));

	// Prepare the progress dialog.
	int nr = displayList.size() ;
	QProgressDialog progress (tr("Averaging images"), tr("Cancel"), 0, height*nr, 0);
	progress.setMinimumDuration(0) ;
	progress.setCancelButtonText (tr("Stop it !")) ;
	progress.setWindowModality(Qt::WindowModal) ;

	// Loop over all images.
	for (i=0 ; i<nr ; i ++ ) 
	{
		// Load the image.
		img.load (displayList[i]->filePath[0]) ;

		// Increment pixel values.
		for (y=0 ; y<height ; y++)
		{
			for (x=0 ; x<width ; x++)
			{
				clr = img.pixel(x,y) ;
 				int pos = x + y * width;
				red[pos] += qRed (clr) ;
				green[pos] += qGreen (clr) ;
				blue[pos] += qBlue (clr) ;
			}

			// Show progression.
			progress.setValue (i*height+y) ;
			if (progress.wasCanceled())
			{
				delete[] red ;
				delete[] green ;
				delete[] blue ;
				spnMkr->setEnabled (true) ;
				btnAlign->setEnabled (true) ;
				btnAverage->setEnabled (true) ;
				btnRemove->setEnabled (true) ;
				btnRevert->setEnabled (true) ;
				btnQuit->setEnabled (true) ;
				return ;
			}
			QCoreApplication::instance()->processEvents();
		}
	}

	// Divide the results by the number of images.
	for (i=0 ; i<width * height ; i++ )
	{
		red[i] /= nr ;
		green[i] /= nr ;
		blue[i] /= nr ;
	}

	// Draw the data into the new image.
	for (y=0 ; y<height ; y++)
		for (x=0 ; x<width ; x++)
        {
 			int pos = x+y*width;
			img.setPixel (x,y, qRgb(red[pos],green[pos],blue[pos])) ;
        }

    // Close the progress dialog.
    progress.reset() ;

	// Show the image.
	QNImageDialog dlg (this) ;
	dlg.setImage (img) ;
	if ( dlg.exec() == QDialog::Accepted )
	{
		// Save the new image
		QString file = QFileDialog::getSaveFileName(this,
		                        tr("Save final image as..."),
		                        QString (), tr("Images (*.jpg *.png)"));
		if ( ! file.isNull() )
        {
			// Write image to disk
			img.save(file,0, jpgQuality) ;

			// Copy EXIF metadata.
            if ( file.endsWith("jpg",Qt::CaseInsensitive) )
            {
			    ImagePtr imageExifRead = Exiv2::ImageFactory::open(displayList[0]->filePath[0].toStdString());
			    imageExifRead->readMetadata();
			    ImagePtr imageExifWrite = Exiv2::ImageFactory::open(file.toStdString());
			    imageExifWrite->setExifData(imageExifRead->exifData());
		    	imageExifWrite->writeMetadata();
			}
		}
	}

	// Free the data.
	delete[] red ;
	delete[] green ;
	delete[] blue ;

	// Enable previously disabled actions.
	spnMkr->setEnabled (true) ;
	btnAlign->setEnabled (true) ;
	btnAverage->setEnabled (true) ;
	btnRemove->setEnabled (true) ;
	btnRevert->setEnabled (true) ;
	btnQuit->setEnabled (true) ;
}




void ImmixWin :: SwitchTo (int idx)
{
	if ( activeIndex != 0 )
		displayList[activeIndex]->Unload () ;
	displayList[idx]->Load() ;
	stack->setCurrentIndex (idx) ;
	activeIndex = idx ;

    // Show file name in label.
    labFile->setText ( QFileInfo(displayList[idx]->filePath[0]).fileName() ) ;

    // Give focus to slider.
    slider->setFocus (Qt::OtherFocusReason) ;
}


void ImmixWin::Revert (void)
{
	// Check if this is really the desired action.
	if (QMessageBox::question (this, tr("Revert"), 
		   tr("Revert to unaligned images?\nThis will delete the aligned images."), 
		   QMessageBox::Yes|QMessageBox::No, QMessageBox::No) 
		 & (QMessageBox::No|QMessageBox::Cancel) )
		return ;
	
	// Loop over images.
	foreach (QPointer<ImmixDisplay> display, displayList)
		display->RevertFile () ;		
	
	// Move markers (as image size had been increased)
	mkr1 -= QPoint(MkrSz,MkrSz) ;
	mkr2 -= QPoint(MkrSz,MkrSz) ;

    // Give input focus to slider.
    slider->setFocus (Qt::OtherFocusReason) ;

	update() ;	
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                         DISPLAY WINDOW 
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


ImmixDisplay::ImmixDisplay (ImmixWin* parent, QString fname)
	: QWidget (parent)
{
	filePath.prepend (fname) ;
	winPtr = parent ;
	imgPtr = 0 ;
	currentZoomLevel = 1. ;
    setCursor (Qt::CrossCursor) ;
}

void ImmixDisplay::Load (void)
{
	if ( imgPtr == 0 )
		imgPtr = new QImage (filePath[0]) ;
}

void ImmixDisplay::Unload (void)
{
	if ( imgPtr )
		delete imgPtr ;
	imgPtr = 0 ;
}

void ImmixDisplay::ChangeFile (QString fname)
{
	filePath.prepend (fname) ;
	if ( imgPtr )
	{
		delete imgPtr ;
		imgPtr = new QImage (filePath[0]) ;
	}
}

void ImmixDisplay::RevertFile ()
{
	// If we're at the last file.
    if ( filePath.size() <= 1 )
    {
		QMessageBox::critical(this, "immix", tr("The displayed files are the original ones !") ) ;
        return ;
    }

    // Delete the file.
    QFile::remove (filePath[0]) ;

    // Remove the first file from the histoy.
	filePath.removeFirst () ;

    // Load the old image.
	if ( imgPtr )
	{
		delete imgPtr ;
		imgPtr = new QImage (filePath[0]) ;
	}
}

void ImmixDisplay::paintEvent (QPaintEvent*)
{
	QPainter painter (this) ;
    painter.setBrush (QColor("white")) ;
    painter.drawRect (painter.window()) ;

	if ( imgPtr )
	{
	// Calculate geometry for painting the whole image.
		QRect paintRect = painter.window().
                          adjusted(margin,margin,-margin,-margin*2-int(ImmixWin::ZoomFactor*ImmixWin::MkrSz)) ;
		int offsetX = 0, offsetY = 0 ;
		float zoomX = paintRect.width() / float(imgPtr->width()) ;
		float zoomY = paintRect.height() / float(imgPtr->height()) ;
		if ( zoomX < zoomY )
		{	
			currentZoomLevel = zoomX ;
			offsetY = int( ( paintRect.height() - imgPtr->height()*zoomX )/2.) ;
		}
		else
		{	
			currentZoomLevel = zoomY ;
			offsetX = int( ( paintRect.width() - imgPtr->width()*zoomY )/2.) ;
		}

		QImage toDraw = imgPtr->scaled(paintRect.size(), 
		                                      Qt::KeepAspectRatio,
		                                      Qt::SmoothTransformation) ;
		QPainter imgPainter (&toDraw) ;
		if ( winPtr->mkr1.x() != -1 )
		{
			imgPainter.setPen(QColor("#FF8000")) ;
			imgPainter.drawRect (int( currentZoomLevel*(winPtr->mkr1.x()-ImmixWin::MkrSz/2) ),
			                     int( currentZoomLevel*(winPtr->mkr1.y()-ImmixWin::MkrSz/2) ), 
			                     int( currentZoomLevel*ImmixWin::MkrSz ),
			                     int( currentZoomLevel*ImmixWin::MkrSz ) ) ;
		}
		if ( winPtr->mkr2.x() != -1 )
		{
			imgPainter.setPen(QColor("#80FF00")) ;
			imgPainter.drawRect (int( currentZoomLevel*(winPtr->mkr2.x()-ImmixWin::MkrSz/2) ),
			                     int( currentZoomLevel*(winPtr->mkr2.y()-ImmixWin::MkrSz/2) ), 
			                     int( currentZoomLevel*ImmixWin::MkrSz ),
			                     int( currentZoomLevel*ImmixWin::MkrSz ) ) ;
		}
		painter.drawImage (margin,margin,toDraw) ;

		// Draw selected zones.
		if ( winPtr->mkr1.x() != -1 )
		{
			QRect refRect1 (winPtr->mkr1.x()-ImmixWin::MkrSz/2,winPtr->mkr1.y()-ImmixWin::MkrSz/2,ImmixWin::MkrSz,ImmixWin::MkrSz) ;
			QImage refImg1 = imgPtr->copy(refRect1).scaled(int(ImmixWin::ZoomFactor*ImmixWin::MkrSz),int(ImmixWin::ZoomFactor*ImmixWin::MkrSz)) ;

            painter.setPen(QColor("#FF8000")) ;
            painter.drawRect(margin-1, paintRect.bottom()+margin-1, refImg1.width()+1, refImg1.height()+1) ;
			painter.drawImage(margin, paintRect.bottom()+margin, refImg1) ;
		}
		if ( winPtr->mkr2.x() != -1 )
		{
			QRect refRect2 (winPtr->mkr2.x()-ImmixWin::MkrSz/2, winPtr->mkr2.y()-ImmixWin::MkrSz/2, ImmixWin::MkrSz, ImmixWin::MkrSz) ;
			QImage refImg2 = imgPtr->copy(refRect2).scaled(int(ImmixWin::ZoomFactor*ImmixWin::MkrSz),int(ImmixWin::ZoomFactor*ImmixWin::MkrSz)) ;

			painter.setPen(QColor("#80FF00")) ;
            painter.drawRect(int(ImmixWin::ZoomFactor*ImmixWin::MkrSz)+2*margin-1, paintRect.bottom()+margin-1, refImg2.width()+1, refImg2.height()+1) ;
			painter.drawImage(int(ImmixWin::ZoomFactor*ImmixWin::MkrSz)+2*margin, paintRect.bottom()+margin, refImg2) ;
		}
	}
	else
		cout << "Attention, pas d'image à dessiner!!!" << endl ;
}


 void ImmixDisplay::mouseReleaseEvent(QMouseEvent* ev)
{
	// Modify marker 2 with right button, marker 1 with any other button click.
	QPoint* mkrPtr = ev->button() == Qt::RightButton ? &( winPtr->mkr2 ) : & (winPtr->mkr1) ;

	// Change marker position
	*mkrPtr = (ev->pos()-QPoint(margin, margin)) / currentZoomLevel ;
	update () ;
	winPtr->btnAlign->setEnabled ( (winPtr->mkr1.x()!=-1) && (winPtr->mkr2.x()!=-1) ) ;
}









